"""
This module contains the tasks that initialize quantities in the state and parameters
objects.

These are typically used in the initialization recipe of the Algorithm object.
"""

import logging
import numpy as np
import copy
from qclab import functions, Simulation

logger = logging.getLogger(__name__)


def initialize_variable_objects(sim: Simulation, state: dict, parameters: dict):
    """
    Populates the State object with non-private variables in ``sim.initial_state``, and an empty
    dictionary for storing output quantities.

    For any non-private (i.e., not beginning with "_") ndarray in ``sim.initial_state``,
    a new array is created in ``state`` with shape ``(batch_size, *original_shape)`` where
    ``original_shape`` is the shape of the array in ``sim.initial_state``.
    The new array is initialized by copying the original array into each slice along the first axis.

    Reads
    -----
    sim.initial_state[name]: ndarray of shape original_shape with dtype=original_dtype
        Any ndarray in initial_state that does not have a key beginning with "_".

    Writes
    ------
    state[name]: ndarray of shape (B, *original_shape) and dtype=original_dtype
        Corresponding ndarray with a shape expanded over a new trajecgtory index.
    state["output_dict"]: dict
        Dictionary to store output quantities during a simulation.

    Notes
    -----
    * B = sim.settings.batch_size
    """
    for name in sim.initial_state.keys():
        obj = sim.initial_state[name]
        if isinstance(obj, np.ndarray) and name[0] != "_":
            initial_shape = np.shape(obj)
            new_obj = np.ascontiguousarray(
                np.zeros((sim.settings.batch_size, *initial_shape), dtype=obj.dtype)
                + obj[np.newaxis]
            )
            logger.info(
                "Initializing state variable %s with shape %s.", name, new_obj.shape
            )
            state[name] = new_obj
        elif name[0] != "_":
            logger.warning(
                "Variable %s in sim.initial_state is not a numpy.ndarray, "
                "skipping initialization in state variable object.",
                name,
            )
    state["output_dict"] = {}
    return state, parameters


def initialize_norm_factor(
    sim: Simulation,
    state: dict,
    parameters: dict,
    norm_factor_name: str = "norm_factor",
):
    """
    Assigns the normalization factor to the State object.

    When collected values are summed in the Data object the normalization factor
    is used to normalize the sum to a trajectory average. In all algorithms in
    QC Lab this is equivalent to the batch size.

    Optional Keyword Arguments
    --------------------------
    norm_factor_name:
        Name of the normalization factor in the State object.

    Writes
    ------
    state[norm_factor_name]:
        Normalization factor for trajectory averages.
    """
    state[norm_factor_name] = sim.settings.batch_size
    return state, parameters


def initialize_branch_seeds(
    sim: Simulation,
    state: dict,
    parameters: dict,
    seed_name: str = "seed",
    branch_ind_name: str = "branch_ind",
):
    """
    Distributes random seeds over branches for deterministic surface hopping.

    This is done by first assuming that the number of branches is equal to the number of quantum states.
    Then, a branch index is created which gives the branch index of each seed in the batch.
    Then a new set of seeds is created by floor dividing the original seeds by the number
    of branches so that the seeds corresponding to different branches within the same
    trajectory are the same.

    Notably, this leads to the number of unique classical initial conditions
    being equal to the number of trajectories divided by the number of
    branches in deterministic surface hopping.

    Optional Keyword Arguments
    --------------------------
    seed_name:
        Name of seed array in State object.
    branch_ind_name:
        Name of the branch index array in ``state``.

    Reads
    -----
    state[seed_name]: ndarray of shape (B,), dtype=int
        Seed for each trajectory.

    Writes
    ------
    state[seed_name]
        Seed for each trajectory remapped for the number of branches.
    state[branch_ind_name]
        Branch index for each trajectory.

    Notes
    -----
    * B = sim.settings.batch_size
    """
    # First ensure that the number of branches is correct.
    if sim.algorithm.settings.fssh_deterministic:
        num_branches = sim.model.constants.num_quantum_states
    else:
        num_branches = 1
    batch_size = sim.settings.batch_size

    if batch_size % num_branches != 0:
        logger.error(
            "Batch size must be an integer multiple of sim.model.constants.num_quantum_states"
        )
        raise ValueError(
            "Batch size must be an integer multiple of sim.model.constants.num_quantum_states"
        )

    # Next, determine the number of trajectories that have been run by assuming that
    # the minimum seed in the current batch of seeds is the number of trajectories
    # that have been run modulo num_branches.
    orig_seed = state[seed_name]
    # Now construct a branch index for each trajectory in the expanded batch.
    state[branch_ind_name] = np.tile(
        np.arange(num_branches), batch_size // num_branches
    )
    # Now generate the new seeds for each trajectory in the expanded batch.
    new_seeds = orig_seed // num_branches
    state[seed_name] = new_seeds
    return state, parameters


def initialize_z_mcmc(
    sim: Simulation,
    state: dict,
    parameters: dict,
    seed_name: str = "seed",
    z_name: str = "z",
):
    """
    Initializes :ref:`complex classical coordinates <coordinates>` according to Boltzmann statistics using Markov
    Chain Monte Carlo with a Metropolis-Hastings algorithm.

    The algorithm has two modes, separable and non-separable. In the separable
    mode, each classical coordinate is evolved as an independent random walker.
    In the non-separable mode, the full classical coordinate vector is evolved as a
    single random walker. The separable mode converges much faster but assumes that
    the classical Hamiltonian can be written as a sum of independent terms depending
    on each classical coordinate.

    Optional Keyword Arguments
    --------------------------
    seed_name:
        Name of seed array in State object.
    z_name:
        Name of destination attribute in ``state``.

    Constants and Settings
    ----------------------
    sim.model.constants.mcmc_burn_in_size: int, default: 1000
        Burn-in step count.
    sim.model.constants.mcmc_sample_size: int, default: 10000
        Number of retained samples.
    sim.model.constants.mcmc_std: float, default: 1.0
        Sampling standard deviation.
    sim.model.constants.mcmc_init_z: ndarray, default: random sample generated by ``functions.gen_sample_gaussian``
        Initial value of the coordinate at the outset of sampling.
    sim.model.constants.kBT: float
        Thermal energy.

    Ingredients
    -----------
    h_c:
        Classical Hamiltonian.

    Reads
    -----
    state[seed_name]: ndarray of shape (B,), dtype=int
        Seed for each trajectory.

    Writes
    ------
    state[z_name]: ndarray of shape (B, C), dtype=complex128
        Classical coordinate in each trajectory.

    Notes
    -----
    * B = sim.settings.batch_size
    * C = sim.model.constants.num_classical_coordinates
    """
    seed = state[seed_name]
    burn_in_size = sim.model.constants.get("mcmc_burn_in_size", 1000)
    sample_size = sim.model.constants.get("mcmc_sample_size", 10000)
    mcmc_h_c_separable = sim.model.constants.get("mcmc_h_c_separable", True)
    burn_in_seeds = np.arange(burn_in_size)
    sample_seeds = np.arange(sample_size)
    save_inds = np.zeros(len(seed), dtype=int)
    out_tmp = np.zeros(
        (sample_size, sim.model.constants.num_classical_coordinates), dtype=complex
    )
    for s, seed_s in enumerate(seed):
        np.random.seed(seed_s)
        save_inds[s] = np.random.randint(0, sample_size)
    mcmc_z_initial, _ = functions.gen_sample_gaussian(
        sim.model.constants, z_initial=None, seed=0, separable=False
    )
    sample = sim.model.constants.get("mcmc_initial_z", mcmc_z_initial)
    h_c, _ = sim.model.get("h_c")
    if mcmc_h_c_separable:
        for s, seed_s in enumerate(burn_in_seeds):
            previous_sample = np.copy(sample)
            previous_z = np.diag(previous_sample)
            previous_energy = h_c(
                sim.model, parameters, z=previous_z, batch_size=len(previous_z)
            )
            proposed_sample, rand = functions.gen_sample_gaussian(
                sim.model.constants,
                z_initial=previous_sample,
                seed=seed_s,
                separable=True,
            )
            new_z = np.diag(proposed_sample)
            new_energy = h_c(sim.model, parameters, z=new_z, batch_size=len(new_z))
            thresh = np.minimum(
                np.ones(sim.model.constants.num_classical_coordinates),
                np.exp(-(new_energy - previous_energy) / sim.model.constants.kBT),
            )
            sample[rand < thresh] = proposed_sample[rand < thresh]
        for s, seed_s in enumerate(sample_seeds):
            previous_sample = np.copy(sample)
            previous_z = np.diag(previous_sample)
            previous_energy = h_c(
                sim.model, parameters, z=previous_z, batch_size=len(previous_z)
            )
            proposed_sample, rand = functions.gen_sample_gaussian(
                sim.model.constants,
                z_initial=previous_sample,
                seed=seed_s,
                separable=True,
            )
            new_z = np.diag(proposed_sample)
            new_energy = h_c(sim.model, parameters, z=new_z, batch_size=len(new_z))
            thresh = np.minimum(
                np.ones(sim.model.constants.num_classical_coordinates),
                np.exp(-(new_energy - previous_energy) / sim.model.constants.kBT),
            )
            sample[rand < thresh] = proposed_sample[rand < thresh]
            out_tmp[s] = sample
            state[z_name] = out_tmp[save_inds]
    else:
        # If not separable, do the full MCMC.
        for s, seed_s in enumerate(burn_in_seeds):
            previous_sample = np.copy(sample)
            previous_energy = h_c(
                sim.model,
                parameters,
                z=previous_sample,
                batch_size=len(previous_sample),
            )
            proposed_sample, rand = functions.gen_sample_gaussian(
                sim.model.constants,
                z_initial=previous_sample,
                seed=seed_s,
                separable=False,
            )
            new_energy = h_c(
                sim.model,
                parameters,
                z=proposed_sample,
                batch_size=len(proposed_sample),
            )
            thresh = min(
                1, np.exp(-(new_energy - previous_energy) / sim.model.constants.kBT)
            )
            if rand < thresh:
                sample = proposed_sample
        for s, seed_s in enumerate(sample_seeds):
            previous_sample = np.copy(sample)
            previous_energy = h_c(
                sim.model,
                parameters,
                z=previous_sample,
                batch_size=len(previous_sample),
            )
            proposed_sample, rand = functions.gen_sample_gaussian(
                sim.model.constants,
                z_initial=previous_sample,
                seed=seed_s,
                separable=False,
            )
            new_energy = h_c(
                sim.model,
                parameters,
                z=proposed_sample,
                batch_size=len(proposed_sample),
            )
            thresh = min(
                1, np.exp(-(new_energy - previous_energy) / sim.model.constants.kBT)
            )
            if rand < thresh:
                sample = proposed_sample
            out_tmp[s] = sample
        state[z_name] = out_tmp[save_inds]
    return state, parameters


def initialize_z(
    sim: Simulation,
    state: dict,
    parameters: dict,
    seed_name: str = "seed",
    z_name: str = "z",
):
    """
    Initializes the classical coordinate by using the init_classical function from the
    Model object.

    Optional Keyword Arguments
    --------------------------
    seed_name:
        Name of seed array in State object.
    z_name:
        Name of classical coordinate in State object.

    Ingredients
    -----------
    init_classical:
        Classical coordinate initialization.

    Reads
    -----
    state[seed_name]: ndarray of shape (B,), dtype=int
        Seed in each trajectory.

    Writes
    ------
    state[z_name] : ndarray of shape (B, C), dtype=complex128
        Initialized classical coordinates.

    Notes
    -----
    * B = sim.settings.batch_size
    * C = sim.model.constants.num_classical_coordinates
    """
    seed = state[seed_name]
    init_classical, has_init_classical = sim.model.get("init_classical")
    if has_init_classical:
        state[z_name] = init_classical(sim.model, parameters, seed=seed)
    else:
        state, parameters = initialize_z_mcmc(
            sim, state, parameters, seed_name=seed_name, z_name=z_name
        )
    return state, parameters


def copy_in_state(
    sim: Simulation, state: dict, parameters: dict, orig_name: str, copy_name: str
):
    """
    Creates a copy of a variable in the State object with a new name.


    Optional Keyword Arguments
    --------------------------
    orig_name:
        Name of the original in the State object.
    copy_name:
        Name of the copy in the State object.

    Reads
    -----
    state[orig_name]:
        Object to be copied.

    Writes
    ------
    state[copy_name]:
        Copy of object.
    """
    if orig_name in state.keys():
        state[copy_name] = copy.copy(state[orig_name])
    return state, parameters


def copy_in_parameters(
    sim: Simulation, state: dict, parameters: dict, orig_name: str, copy_name: str
):
    """
    Creates a copy of a variable in the Parameters object with a new name.


    Optional Keyword Arguments
    --------------------------
    orig_name:
        Name of the original in the Parameters object.
    copy_name:
        Name of the copy in the Parameters object.

    Reads
    -----
    parameters[orig_name]:
        Object to be copied.

    Writes
    ------
    parameters[copy_name]:
        Copy of object.
    """
    parameters[copy_name] = copy.copy(parameters[orig_name])
    return state, parameters


def copy_to_parameters(
    sim: Simulation,
    state: dict,
    parameters: dict,
    state_name: str,
    parameters_name: str,
):
    """
    Copies an object from the State object to the Parameters object.

    Optional Keyword Arguments
    --------------------------
    state_name:
        Name of the object in the State object.
    parameters_name:
        Name of the object in the Parameters object.

    Reads
    -----
    state[state_name]:
        Object to be copied.

    Writes
    ------
    parameters[parameters_name]:
        Copy of object.
    """
    parameters[parameters_name] = copy.copy(state[state_name])
    return state, parameters


def initialize_active_surface(
    sim: Simulation,
    state: dict,
    parameters: dict,
    act_surf_ind_0_name: str = "act_surf_ind_0",
    act_surf_ind_name: str = "act_surf_ind",
    act_surf_name: str = "act_surf",
    init_act_surf_rand_vals_name: str = "init_act_surf_rand_vals",
    wf_adb_name: str = "wf_adb",
):
    """
    Initializes the active surface, active surface index, and initial active
    surface index for FSSH.

    If ``fssh_deterministic=True`` it will set the initial active surface index
    to be the same as the branch index and assert that the number of branches is
    equal to the number of quantum states.

    If ``fssh_deterministic=False`` it will stochastically sample the active
    surface from the density corresponding to the initial quantum wavefunction
    in the adiabatic basis.

    Optional Keyword Arguments
    --------------------------
    act_surf_ind_0_name:
        Name of the initial active surface index in the State object.
    act_surf_ind_name:
        Name of the active surface index in the State object.
    act_surf_name:
        Name of the active surface in the State object.
    init_act_surf_rand_vals_name: str, default:
        Name of the random numbers for active surface initialization in FSSH.
    wf_adb_name: str, default:
        Name of the adiabatic wavefunction in the State object.

    Constants and Settings
    ----------------------
    sim.algorithm.settings.fssh_deterministic: Bool
        Boolean indicating if the FSSH simulation is deterministic.

    Reads
    -----
    state[wf_adb_name]: ndarray of shape (B, N), dtype=complex128
        Wavefunction coefficients in the adiabatic basis.

    Writes
    ------
    state[act_surf_ind_0_name]: ndarray of shape (B,), dtype=int
        Active surface index at the start of the simulation.
    state[act_surf_ind_name]: ndarray of shape (B,), dtype=int
        Active surface index.
    state[act_surf_name]: ndarray of shape (B, N), dtype=int
        Active surface vector in adiabatic basis: 1 if active 0 if not.

    Notes
    -----
    * B = sim.settings.batch_size
    * N = sim.model.constants.num_quantum_states
    """
    wf_adb = state[wf_adb_name]
    init_act_surf_rand_vals = state[init_act_surf_rand_vals_name]
    if sim.algorithm.settings.fssh_deterministic:
        num_branches = sim.model.constants.num_quantum_states
    else:
        num_branches = 1
    num_states = sim.model.constants.num_quantum_states
    num_trajs = sim.settings.batch_size // num_branches
    if sim.algorithm.settings.fssh_deterministic:
        act_surf_ind_0 = np.tile(np.arange(num_branches, dtype=int), (num_trajs, 1))
    else:
        cumulative_pops = np.cumsum(
            np.real(np.abs(wf_adb.reshape((num_trajs, num_branches, num_states))) ** 2),
            axis=-1,
        )
        act_surf_ind_0 = np.argmax(
            cumulative_pops > init_act_surf_rand_vals[:, :, np.newaxis], axis=-1
        ).astype(int)
    act_surf_ind_0 = act_surf_ind_0.reshape(-1)
    state[act_surf_ind_0_name] = np.copy(act_surf_ind_0)
    state[act_surf_ind_name] = np.copy(act_surf_ind_0)
    act_surf = np.zeros((num_trajs * num_branches, num_states), dtype=int)
    traj_inds = np.repeat(np.arange(num_trajs), num_branches)
    branch_inds = np.tile(np.arange(num_branches), num_trajs)
    traj_branch_ind = traj_inds * num_branches + branch_inds
    act_surf[traj_branch_ind, act_surf_ind_0] = 1
    state[act_surf_name] = act_surf
    return state, parameters


def initialize_random_values_fssh(
    sim: Simulation,
    state: dict,
    parameters: dict,
    hop_prob_rand_vals_name: str = "hop_prob_rand_vals",
    init_act_surf_rand_vals_name: str = "init_act_surf_rand_vals",
    seed_name: str = "seed",
):
    """
    Initializes a set of random numbers using the trajectory seeds for FSSH.

    .. rubric:: Required Constants
    None

    Optional Keyword Arguments
    --------------------------
    hop_prob_rand_vals_name:
        Name of the random numbers for hop decisions in the State object.
    init_act_surf_rand_vals_name:
        Name of the random numbers for active surface initialization in FSSH.
    seed_name:
        Name of seed array in State object.

    Constants and Settings
    ----------------------
    sim.algorithm.settings.fssh_deterministic: Bool
        Boolean indicating if the FSSH simulation is deterministic.

    Reads
    -----
    state[seed_name]

    Writes
    ------
    state[hop_prob_rand_vals_name]: ndarray of shape (B//b, t), dtype=float64
        Random numbers for hop decisions.
    state[init_act_surf_rand_vals_name]
        Random numbers for active surface selection in stochastic FSSH.

    Notes
    -----
    * B = sim.settings.batch_size
    * b = sim.model.constants.num_quantum_states if fssh_deterministic == True, b = 1 otherwise.
    * t is the number of update timesteps.
    """
    seed = state[seed_name]
    if sim.algorithm.settings.fssh_deterministic:
        num_branches = sim.model.constants.num_quantum_states
    else:
        num_branches = 1
    num_trajs = sim.settings.batch_size // num_branches
    hop_prob_rand_vals = np.zeros((num_trajs, len(sim.settings.t_update)))
    init_act_surf_rand_vals = np.zeros((num_trajs, num_branches))
    for nt in range(num_trajs):
        np.random.seed(seed[int(nt * num_branches)])
        hop_prob_rand_vals[nt] = np.random.rand(len(sim.settings.t_update))
        init_act_surf_rand_vals[nt] = np.random.rand(num_branches)
    state[hop_prob_rand_vals_name] = hop_prob_rand_vals
    state[init_act_surf_rand_vals_name] = init_act_surf_rand_vals
    return state, parameters


def initialize_dm_adb_0_fssh(
    sim: Simulation,
    state: dict,
    parameters: dict,
    dm_adb_0_name: str = "dm_adb_0",
    wf_adb_name: str = "wf_adb",
):
    """
    Initializes the initial adiabatic density matrix for FSSH.

    Optional Keyword Arguments
    --------------------------
    dm_adb_0_name:
        Name of the initial adiabatic density matrix in the State object.
    wf_adb_name:
        Name of the adiabatic wavefunction in the State object.

    Reads
    -----
    state[wf_adb_name]: ndarray of shape (B, N), dtype=complex128
        Wavefunction coefficients in the adiabatic basis.


    Writes
    ------
    state[dm_adb_0_name]: ndarray of shape (B, N, N), dtype=complex128
        Initial density matrix in the adiabatic basis.


    Notes
    -----
    * B = sim.settings.batch_size
    * N = sim.model.constants.num_quantum_states
    """
    wf_adb = state[wf_adb_name]
    state[dm_adb_0_name] = np.einsum(
        "ti,tj->tij",
        wf_adb,
        np.conj(wf_adb),
        optimize="greedy",
    )
    return state, parameters



"""
This module contains utility helpers for optional jit compilation and in-memory logging.
"""

import logging
from io import StringIO

DISABLE_NUMBA = False
try:
    import numba as _
except ImportError:
    DISABLE_NUMBA = True

DISABLE_H5PY = False
try:
    import h5py as _
except ImportError:
    DISABLE_H5PY = True

DISABLE_ASE = False
try:
    import ase as _
except ImportError:
    DISABLE_ASE = True


logger = logging.getLogger(__name__)


def qc_lab_custom_jit(func=None, **kwargs):
    """
    Dummy jit decorator that does nothing if numba is not available.
    """

    def decorator(func):
        return func

    if func is None:
        return decorator
    return decorator(func)


def qc_lab_custom_njit(func=None, **kwargs):
    """
    Dummy njit decorator that does nothing if numba is not available.
    """

    def decorator(func):
        return func

    if func is None:
        return decorator
    else:
        return decorator(func)


if DISABLE_NUMBA:
    logger.info(
        "Numba is disabled; using dummy jit decorator.",
    )
    njit = qc_lab_custom_njit
    jit = qc_lab_custom_jit
else:
    try:
        from numba import jit, njit
    except ImportError:
        jit = qc_lab_custom_jit
        njit = qc_lab_custom_njit


_log_stream = StringIO()


class QCDataHandler(logging.Handler):
    """
    Logging handler that stores logs in memory.
    """

    def emit(self, record):
        msg = self.format(record)
        _log_stream.write(msg + "\n")


def configure_memory_logger(level=logging.INFO):
    """
    Configure root logger to store logs without printing.
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    for handler in list(root_logger.handlers):
        root_logger.removeHandler(handler)
    handler = QCDataHandler()
    formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)
    return handler


def get_log_output() -> str:
    """
    Return all collected log messages.
    """
    return _log_stream.getvalue()


def reset_log_output() -> None:
    """
    Clear all collected log messages.

    Resets the in-memory log stream so that subsequent calls to
    :func:`get_log_output` return only the messages emitted after this
    function is called.
    """
    _log_stream.seek(0)
    _log_stream.truncate(0)


__all__ = [
    "jit",
    "njit",
    "configure_memory_logger",
    "get_log_output",
    "reset_log_output",
    "DISABLE_NUMBA",
    "DISABLE_H5PY",
]



"""
This module imports the dynamics drivers to qclab.dynamics.
"""

from qclab.dynamics.parallel_driver_multiprocessing import (
    parallel_driver_multiprocessing,
)
from qclab.dynamics.parallel_driver_mpi import parallel_driver_mpi
from qclab.dynamics.serial_driver import serial_driver
from qclab.dynamics.dynamics import run_dynamics



"""
This module tests the serial and multiprocessing drivers for a few simple cases.
"""

import pytest


def test_drivers_spinboson():
    import numpy as np
    from qclab import Simulation  # import simulation class
    from qclab.models import SpinBoson  # import model class
    from qclab.algorithms import MeanField  # import algorithm class
    from qclab.dynamics import (
        serial_driver,
        parallel_driver_multiprocessing,
    )  # import dynamics driver

    sim = Simulation()
    sim.settings.progress_bar = False
    sim.settings.num_trajs = 200
    sim.settings.batch_size = 50
    sim.settings.tmax = 10
    sim.settings.dt_update = 0.01

    sim.model = SpinBoson(
        {
            "V": 0.5,
            "E": 0.5,
            "A": 100,
            "W": 0.1,
            "l_reorg": 0.005,
            "boson_mass": 1.0,
            "kBT": 1.0,
        }
    )
    sim.algorithm = MeanField()
    sim.model.initialize_constants()
    sim.initial_state["wf_db"] = np.zeros(
        (sim.model.constants.num_quantum_states), dtype=complex
    )
    sim.initial_state["wf_db"][0] += 1.0
    print("Running serial driver...")
    data_serial = serial_driver(sim)
    print("Running parallel multiprocessing driver...")
    data_parallel_multiprocessing = parallel_driver_multiprocessing(sim)
    print("Comparing results...")
    for key, val in data_serial.data_dict.items():
        if isinstance(val, np.ndarray):
            assert np.allclose(val, data_parallel_multiprocessing.data_dict[key])
    print("parallel and serial results match!")
    return


@pytest.mark.mpi
def test_drivers_spinboson_mpi():
    """
    This test runs the serial and parallel drivers for the SpinBoson model using MPI. It
    compares the results to ensure they match.

    This test requires MPI to be set up and run with a command like:
    mpirun -n 4 pytest -m mpi -s tests/test_drivers.py
    """
    import numpy as np
    from qclab import Simulation  # import simulation class
    from qclab.models import SpinBoson  # import model class
    from qclab.algorithms import MeanField  # import algorithm class

    pytest.importorskip("mpi4py")
    from mpi4py import MPI  # import MPI for parallel processing
    from qclab.dynamics import (
        serial_driver,
        parallel_driver_mpi,
    )  # import dynamics driver

    sim = Simulation()
    sim.settings.progress_bar = False
    sim.settings.num_trajs = 200
    sim.settings.batch_size = 50
    sim.settings.tmax = 10
    sim.settings.dt_update = 0.01

    sim.model = SpinBoson(
        {
            "V": 0.5,
            "E": 0.5,
            "A": 100,
            "W": 0.1,
            "l_reorg": 0.005,
            "boson_mass": 1.0,
            "kBT": 1.0,
        }
    )
    sim.algorithm = MeanField()
    sim.model.initialize_constants()
    sim.initial_state["wf_db"] = np.zeros(
        (sim.model.constants.num_quantum_states), dtype=complex
    )
    sim.initial_state["wf_db"][0] += 1.0

    data_parallel_mpi = parallel_driver_mpi(sim)
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    print("finished running parallel driver on rank", rank)
    if rank == 0:
        print("Running serial driver... on rank", rank)
        data_serial = serial_driver(sim)
        print("Comparing results...")
        for key, val in data_serial.data_dict.items():
            if isinstance(val, np.ndarray):
                if key == "seed":
                    # Need to sort seeds since batches are returned in different orders.
                    data_parallel_mpi.data_dict[key] = np.sort(
                        data_parallel_mpi.data_dict[key]
                    )
                assert np.allclose(val, data_parallel_mpi.data_dict[key])
        print("parallel and serial results match!")
    return


def test_incommensurate_batch_size_serial():
    """
    This test checks that the drivers work correctly when the number of trajectories
    is not an integer multiple of the batch size.
    """
    import numpy as np
    from qclab import Simulation  # import simulation class
    from qclab.models import SpinBoson  # import model class
    from qclab.algorithms import MeanField  # import algorithm class
    from qclab.dynamics import serial_driver  # import dynamics driver

    sim = Simulation()
    sim.settings.progress_bar = False
    sim.settings.num_trajs = 200  # Not an integer multiple of batch_size
    sim.settings.batch_size = 50
    sim.settings.tmax = 10
    sim.settings.dt_update = 0.01

    sim.model = SpinBoson(
        {
            "V": 0.5,
            "E": 0.5,
            "A": 100,
            "W": 0.1,
            "l_reorg": 0.005,
            "boson_mass": 1.0,
            "kBT": 1.0,
        }
    )
    sim.algorithm = MeanField()
    sim.model.initialize_constants()
    sim.initial_state["wf_db"] = np.zeros(
        (sim.model.constants.num_quantum_states), dtype=complex
    )
    sim.initial_state["wf_db"][0] += 1.0
    print("Running commensurate batch size serial driver...")
    data_serial = serial_driver(sim)
    sim.settings.batch_size = 30  # Change batch size to not divide num_trajs.
    print("Running incommensurate batch size serial driver...")
    data_serial_incomm = serial_driver(sim)
    print("Comparing results...")
    for key, val in data_serial.data_dict.items():
        if isinstance(val, np.ndarray):
            assert np.allclose(val, data_serial_incomm.data_dict[key])
            # assert np.allclose(val, data_parallel_mpi.data_dict[key])
    print("results match!")
    return


def test_incommensurate_batch_size_multiprocessing():
    """
    This test checks that the drivers work correctly when the number of trajectories
    is not an integer multiple of the batch size.
    """
    import numpy as np
    from qclab import Simulation  # import simulation class
    from qclab.models import SpinBoson  # import model class
    from qclab.algorithms import MeanField  # import algorithm class
    from qclab.dynamics import (
        parallel_driver_multiprocessing,
    )  # import dynamics driver

    sim = Simulation()
    sim.settings.progress_bar = False
    sim.settings.num_trajs = 200  # Not an integer multiple of batch_size
    sim.settings.batch_size = 50
    sim.settings.tmax = 10
    sim.settings.dt_update = 0.01

    sim.model = SpinBoson(
        {
            "V": 0.5,
            "E": 0.5,
            "A": 100,
            "W": 0.1,
            "l_reorg": 0.005,
            "boson_mass": 1.0,
            "kBT": 1.0,
        }
    )
    sim.algorithm = MeanField()
    sim.model.initialize_constants()
    sim.initial_state["wf_db"] = np.zeros(
        (sim.model.constants.num_quantum_states), dtype=complex
    )
    sim.initial_state["wf_db"][0] += 1.0
    print("Running commensurate batch size parallel driver...")
    data_parallel = parallel_driver_multiprocessing(sim)
    sim.settings.batch_size = 30  # Change batch size to not divide num_trajs.
    print("Running incommensurate batch size parallel driver...")
    data_parallel_incomm = parallel_driver_multiprocessing(sim)
    print("Comparing results...")
    for key, val in data_parallel.data_dict.items():
        if isinstance(val, np.ndarray):
            assert np.allclose(val, data_parallel_incomm.data_dict[key])
            # assert np.allclose(val, data_parallel_mpi.data_dict[key])
    print("results match!")
    return


if __name__ == "__main__":
    test_drivers_spinboson()
    test_incommensurate_batch_size_serial()
    test_incommensurate_batch_size_multiprocessing()



"""
This module imports the model classes to qclab.models.
"""

from qclab.models.holstein_lattice import (
    HolsteinLattice,
    HolsteinLatticeReciprocalSpace,
)
from qclab.models.spin_boson import SpinBoson
from qclab.models.fmo_complex import FMOComplex
from qclab.models.tully_problem_one import TullyProblemOne
from qclab.models.tully_problem_two import TullyProblemTwo
from qclab.models.tully_problem_three import TullyProblemThree
from qclab.utils import DISABLE_ASE
if not(DISABLE_ASE):
    from qclab.models.ab_initio import AbInitio



"""
This module contains the Constants class.
"""


class Constants:
    """
    Constants class to define constants and trigger updates when they change.
    """

    def __init__(self, update_function=None):
        self._updating = False
        self._init_complete = False
        self._update_function = update_function

    def __setattr__(self, name, value):
        """
        Overrides attribute setting to call the update function after the attribute is
        changed, preventing recursion.
        """
        super().__setattr__(name, value)
        if not self._updating and name not in {
            "_updating",
            "_update_function",
            "_init_complete",
        }:
            if self._init_complete:
                self._updating = True
                if self._update_function is not None:
                    self._update_function()
                self._updating = False

    def get(self, name, default=None):
        """
        Get the value of an attribute, returning default if it does not exist.

        .. rubric:: Args
        name : str
            The name of the attribute to get.
        default : any, optional
            The default value to return if the attribute does not exist.

        .. rubric:: Returns
        self.name : any
            The value of the attribute.
        """
        return getattr(self, name, default)



"""
This module contains functions used in QC Lab. This includes
functions that are used by tasks, ingredients, and the
dynamics drivers.
"""

import logging
import functools
import numpy as np
from qclab.utils import njit
import qclab.numerical_constants as numerical_constants

logger = logging.getLogger(__name__)


def batch_matvec(mat, vec):
    """
    Multiplies a matrix with a vector where there is an initial set of indices.

    Assumes that the last two indices of ``mat`` are the matrix indices,
    and the last index of ``vec`` is the vector index. The other indices
    must be broadcastable. This rule is needed to ensure that the vector
    is not interpreted as a single matrix.

    (..., i, j) x (..., j) -> (..., i)

    .. rubric:: Args
    mat : ndarray
        Input matrix.
    vec : ndarray
        Input vector.

    .. rubric:: Returns
    out : ndarray
        Result of the multiplication.
    """
    return np.matmul(mat, vec[..., None])[..., 0]


def transform_vec(vec, basis, adb_to_db=False):
    """
    Transforms a vector ``vec`` to  a new basis defined by the
    column vectors of the matrix ``basis``.

    If ``basis`` are eigenvectors of a diabatic Hamiltonian, then this
    transformation ammounts to a transformation from the diabatic
    to adiabatic basis.

    Assumes that the last two indices of ``basis`` are the matrix indices,
    and the last index of ``vec`` is the vector index. The other indices
    must be broadcastable.

    .. rubric:: .. rubric:: Args
    vec : ndarray
        Input vector.
    basis : ndarray
        Basis matrix.
    adb_to_db : bool, optional, default: False
        If True, reverses the direction of the transformation
        by using the Hermitian conjugate of basis.

    .. rubric:: Returns
    out : ndarray
        Transformed matrix
    """
    if adb_to_db:
        out = batch_matvec(basis, vec)
    else:
        out = batch_matvec(np.swapaxes(np.conj(basis), -1, -2), vec)
    return out


def transform_mat(mat, basis, adb_to_db=False):
    """
    Transforms a matrix ``mat`` to a new basis defined by the
    column vectors of the basis ``basis``.

    If ``basis`` are eigenvectors of a diabatic Hamiltonian, then this
    transformation ammounts to a transformation from the diabatic
    to adiabatic basis.

    Assumes that the last two indices of ``mat`` and ``basis`` are the matrix
    indices. The other indices must be broadcastable.

    .. rubric:: Args
    mat : ndarray
        Input matrix.
    basis : ndarray
        Unitary matrix.
    adb_to_db : bool, optional, default: False
        If True, reverses the direction of the transformation
        by using the Hermitian conjugate of basis.

    .. rubric:: Returns
    out : ndarray
        Transformed matrix
    """
    if adb_to_db:
        out = np.matmul(basis, np.matmul(mat, np.swapaxes(basis.conj(), -1, -2)))
    else:
        out = np.matmul(np.swapaxes(basis.conj(), -1, -2), np.matmul(mat, basis))
    return out


@njit
def update_z_rk4_k123_sum(z_k, classical_force, quantum_classical_force, dt_update):
    """
    Low-level function to calculate the intermediate z coordinate and k values
    for RK4 update. Applies to steps 1-3.

    .. rubric:: Args
    z_k : ndarray
        Initial complex coordinate for that update step.
    classical_force : ndarray
        Classical force.
    quantum_classical_force : ndarray
        Quantum-classical force.
    dt_update : float
        Time step for the update.

    .. rubric:: Returns
    out : ndarray
        Updated complex coordinates after applying the RK4 step.
    k : ndarray
        The k value used in the RK4 update.
    """
    batch_size, num_classical_coordinates = z_k.shape
    k = np.empty((batch_size, num_classical_coordinates), dtype=np.complex128)
    out = np.empty((batch_size, num_classical_coordinates), dtype=np.complex128)
    for i in range(z_k.shape[0]):
        for j in range(z_k.shape[1]):
            k[i, j] = -1j * (classical_force[i, j] + quantum_classical_force[i, j])
            out[i, j] = z_k[i, j] + dt_update * k[i, j]
    return out, k


@njit
def update_z_rk4_k4_sum(
    z_0, k1, k2, k3, classical_force, quantum_classical_force, dt_update
):
    """
    Low-level function to calculate the fourth (and final) step for the RK4 update.

    .. rubric:: Args
    z_0 : ndarray
        Initial complex coordinate.
    k1 : ndarray
        First RK4 slope.
    k2 : ndarray
        Second RK4 slope.
    k3 : ndarray
        Third RK4 slope.
    classical_force : ndarray
        Classical force.
    quantum_classical_force : ndarray
        Quantum-classical force.
    dt_update : float
        Time step for the update.

    .. rubric:: Returns
    z_0 : ndarray
        Updated complex coordinates after applying the RK4 step.
    """
    for i in range(z_0.shape[0]):
        for j in range(z_0.shape[1]):
            z_0[i, j] = z_0[i, j] + (dt_update / 6.0) * (
                k1[i, j]
                + 2.0 * k2[i, j]
                + 2.0 * k3[i, j]
                - 1j * (classical_force[i, j] + quantum_classical_force[i, j])
            )
    return z_0


@njit
def dqdp_to_dzc(dq, dp, m, h):
    """
    Convert derivatives w.r.t. q and p (``dq`` and ``dp``, respectively) to
    the derivative w.r.t. zc (``dzc``).

    .. rubric:: Args
    dq : ndarray | None
        Derivative w.r.t. position coordinate.
    dp : ndarray | None
        Derivative w.r.t. momentum coordinate.
    m : ndarray
        classical coordinate mass.
    h : ndarray
        classical coordinate weight.

    .. rubric:: Returns
    dzc : ndarray
        Derivative w.r.t. conjugate complex coordinate.
    """
    dp_present = dp is not None
    dq_present = dq is not None
    if dp_present:
        return 1j * np.sqrt(m * h / 2) * dp
    if dq_present:
        return np.sqrt(0.5 / (m * h)) * dq.astype(np.complex128)
    if dq_present and dp_present:
        return np.sqrt(0.5 / (m * h)) * dq + 1j * np.sqrt(m * h / 2) * dp
    raise ValueError("At least one of dq or dp must be provided.")


@njit
def dzdzc_to_dqdp(dz, dzc, m, h):
    """
    Convert derivatives w.r.t. z and zc (``dz`` and ``dzc``) to derivatives w.r.t.
    q and p (``dq`` and ``dp``). If only one of ``dz`` or ``dzc`` is provided, then
    ``dq`` and ``dp`` are calculated assuming that :math:`dz = dzc^{*}`.

    .. rubric:: Args
    dz : ndarray | None
        Derivative w.r.t. complex z coordinate.
    dzc : ndarray | None
        Derivative w.r.t. conjugate z coordinate.
    m : ndarray
        classical coordinate mass.
    h : ndarray
        classical coordinate weight.

    .. rubric:: Returns
    dq : ndarray
        Derivative w.r.t. position coordinate.
    dp : ndarray
        Derivative w.r.t. momentum coordinate.
    """
    dz_present = dz is not None
    dzc_present = dzc is not None
    if dz_present and dzc_present:
        dq = np.sqrt(0.5 * m * h) * (dz + dzc)
        dp = 1j * np.sqrt(0.5 / (m * h)) * (dz - dzc)
        return dq, dp
    if dz_present:
        dq = np.sqrt(0.5 * m * h) * (dz + np.conj(dz))
        dp = 1j * np.sqrt(0.5 / (m * h)) * (dz - np.conj(dz))
        return dq, dp
    if dzc_present:
        dq = np.sqrt(0.5 * m * h) * (np.conj(dzc) + dzc)
        dp = 1j * np.sqrt(0.5 / (m * h)) * (np.conj(dzc) - dzc)
        return dq, dp
    raise ValueError("At least one of dz or dzc must be provided.")


@njit
def z_to_q(z, m, h):
    """
    Convert complex coordinates to position coordinate.

    .. rubric:: Args
    z : ndarray
        Complex coordinates.
    m : ndarray
        Classical coordinate mass.
    h : ndarray
        Classical coordinate weight.

    .. rubric:: Returns
    q : ndarray
        Position coordinates.
    """
    return np.sqrt(2.0 / (m * h)) * z.real


@njit
def z_to_p(z, m, h):
    """
    Convert complex coordinates to momentum coordinate.

    .. rubric:: Args
    z : ndarray
        Complex coordinates.
    m : ndarray
        Classical coordinate mass.
    h : ndarray
        Classical coordinate weight.

    .. rubric:: Returns
    p : ndarray
        Momentum coordinates.
    """
    return np.sqrt(2.0 * m * h) * z.imag


@njit
def qp_to_z(q, p, m, h):
    """
    Convert real coordinates to complex coordinates.
    If only one of ``q`` or ``p`` is provided, then the other is assumed to be zero.

    .. rubric:: Args
    q : ndarray | None
        Position coordinates.
    p : ndarray | None
        Momentum coordinates.
    m : ndarray
        Classical coordinate mass.
    h : ndarray
        Classical coordinate weight.

    .. rubric:: Returns
    z : ndarray
        Complex coordinates.
    """
    q_present = q is not None
    p_present = p is not None
    if q_present and p_present:
        return np.sqrt(0.5 * m * h) * q + 1j * np.sqrt(0.5 / (m * h)) * p
    if q_present and not p_present:
        return np.sqrt(0.5 * m * h) * q.astype(np.complex128)
    if not q_present and p_present:
        return 1j * np.sqrt(0.5 / (m * h)) * p
    raise ValueError("At least one of q or p must be provided.")


def make_ingredient_sparse(ingredient):
    """
    Wrapper that converts a vectorized ingredient output to a sparse format consisting
    of the indices (``inds``), nonzero elements (``mels``), and ``shape``.

    .. rubric:: Args
    ingredient : function
        Ingredient to be converted to sparse format.

    .. rubric:: Returns
    sparse_ingredient : function
        Sparse version of the ingredient.
    """

    @functools.wraps(ingredient)
    def sparse_ingredient(*args, **kwargs):
        (model, parameters) = args
        out = ingredient(model, parameters, **kwargs)
        inds = np.where(out != 0)
        mels = out[inds]
        shape = np.shape(out)
        return inds, mels, shape

    return sparse_ingredient


def vectorize_ingredient(ingredient):
    """
    Wrapper that vectorizes ingredient functions.
    It assumes that any kwarg is an ndarray that is vectorized over its first
    index. Other kwargs are assumed to not be vectorized.

    .. rubric:: Args
    ingredient : function
        Ingredient to be vectorized.

    .. rubric:: Returns
    vectorized_ingredient : function
        Vectorized version of the ingredient.
    """

    @functools.wraps(ingredient)
    def vectorized_ingredient(*args, **kwargs):
        (model, parameters) = args
        if kwargs.get("z") is not None:
            batch_size = len(kwargs["z"])
        else:
            batch_size = kwargs["batch_size"]
        keys = kwargs.keys()
        kwargs_list = []
        for n in range(batch_size):
            kwargs_n = {"traj_ind": n}
            for key in keys:
                if isinstance(kwargs[key], np.ndarray):
                    kwargs_n[key] = kwargs[key][n]
                else:
                    kwargs_n[key] = kwargs[key]
            kwargs_list.append(kwargs_n)
        out = np.array(
            [ingredient(model, parameters, **kwargs_list[n]) for n in range(batch_size)]
        )
        return out

    return vectorized_ingredient


@njit
def dh_c_dzc_harmonic_jit(z, h, w):
    """
    Derivative of the harmonic oscillator classical Hamiltonian function with respect to
    the conjugate z coordinate.

    .. rubric:: Args
    z : ndarray
        Complex coordinates.
    h : ndarray
        Classical coordinate weight.
    w : ndarray
        Harmonic frequency.

    .. rubric:: Returns
    out : ndarray
        Derivative of the harmonic oscillator classical Hamiltonian function with respect to
        the conjugate z coordinate.
    """

    batch_size, num_classical_coordinates = z.shape
    out = np.empty((batch_size, num_classical_coordinates), dtype=np.complex128)
    w2_over_h = (w**2) / h
    for i in range(batch_size):
        for j in range(num_classical_coordinates):
            zij = z[i, j]
            out[i, j] = complex(w2_over_h[j] * zij.real, h[j] * zij.imag)

    return out


@njit
def h_qc_diagonal_linear_jit(z, gamma):
    """
    Low-level function to generate the diagonal linear quantum-classical Hamiltonian.

    :math:`H_{nm} = \\delta_{nm}\\sum_{j} \\gamma_{nj} (z_{j} + z_{j}^*)`

    .. rubric:: Args
    z : ndarray
        Complex coordinates.
    gamma : ndarray
        Classical coordinate coupling strengths.

    .. rubric:: Returns
    h_qc : ndarray
        Diagonal linear quantum-classical Hamiltonian.
    """
    batch_size = z.shape[0]
    num_classical_coordinates = z.shape[1]
    num_sites = gamma.shape[0]
    h_qc = np.zeros((batch_size, num_sites, num_sites), dtype=np.complex128)
    for b in range(batch_size):
        for i in range(num_sites):
            acc = 0.0
            for j in range(num_classical_coordinates):
                acc += gamma[i, j] * 2.0 * z[b, j].real
            h_qc[b, i, i] = acc
    return h_qc


def gen_sample_gaussian(constants, z_initial=None, seed=None, separable=True):
    """
    Generates a complex number sampled from a Gaussian distribution.

    If ``z_initial`` is provided, then a Gaussian distribution centered around
    ``z_initial`` is sampled.
    If ``z_initial`` is not provided, then a Gaussian distribution centered around
    the origin is sampled.

    If ``separable`` is ``True``, then a different random number is generated
    for each classical coordinate (i.e., each coordinate corresponds to
    a different random walker). If ``False``, then the same random number
    is generated for all classical coordinates (i.e., a single random walker).

    .. rubric:: Args
    constants : Constants
        Constants object.
    z_initial : ndarray | None, default: None
        Center of the Gaussian distribution. If ``None``, the distribution is
        centered around the origin.
    seed : int | None
        Random seed for reproducibility.
    separable : bool, default: True
        Whether to generate a different random number for each classical coordinate.

    .. rubric:: Model Constants
    mcmc_std : float, default: 1.0
        Standard deviation of the Gaussian distribution.

    .. rubric:: Returns
    z : ndarray
        Complex number sampled from a Gaussian distribution.
    rand : ndarray
        Random number(s) used to generate the complex number.
    """
    if seed is not None:
        np.random.seed(seed)
    num_classical_coordinates = constants.num_classical_coordinates
    if separable:
        rand = np.random.rand(num_classical_coordinates)
    else:
        rand = np.random.rand()
    if z_initial is None:
        z_initial = np.zeros(num_classical_coordinates, dtype=complex)
    mcmc_std = constants.get("mcmc_std", 1.0)
    z_re = np.random.normal(
        loc=z_initial.real, scale=mcmc_std, size=num_classical_coordinates
    )
    z_im = np.random.normal(
        loc=z_initial.imag, scale=mcmc_std, size=num_classical_coordinates
    )
    z = z_re + 1j * z_im
    return z, rand


@njit
def calc_sparse_inner_product(inds, mels, shape, vec_l_conj, vec_r, out=None):
    """
    Take a sparse gradient matrix with shape ``(batch_size, num_classical_coordinates,
    num_quantum_state, num_quantum_states)`` and calculate the matrix element of
    the vectors ``vec_l_conj`` and ``vec_r`` with  shape ``(batch_size*num_quantum_states)``.

    .. rubric:: Args
    inds : tuple of ndarrays
        Indices of the nonzero elements in the sparse matrix.
    mels : ndarray
        Nonzero elements of the sparse matrix.
    shape : tuple
        Shape of the sparse matrix.
    vec_l_conj : ndarray
        Left vector (conjugated) for the inner product.
    vec_r : ndarray
        Right vector for the inner product.
    out : ndarray | None
        Preallocated output array. If ``None``, a new array is created.

    .. rubric:: Returns
    out : ndarray
        Result of the inner product with shape ``(batch_size, num_classical_coordinates)``.
    """
    batch_size, num_classical_coordinates, num_quantum_states = (
        shape[0],
        shape[1],
        shape[2],
    )
    r = inds[0]
    c = inds[1]
    a = inds[2]
    b = inds[3]

    if out is None:
        out = np.zeros(batch_size * num_classical_coordinates, dtype=np.complex128)
    if out is not None:
        out.fill(0.0j)

    l_flat = vec_l_conj.reshape(batch_size * num_quantum_states)
    r_flat = vec_r.reshape(batch_size * num_quantum_states)

    for i in range(mels.shape[0]):
        ri = r[i]
        idx_out = ri * num_classical_coordinates + c[i]
        idx_l = ri * num_quantum_states + a[i]
        idx_r = ri * num_quantum_states + b[i]
        out[idx_out] += l_flat[idx_l] * mels[i] * r_flat[idx_r]

    return out


def analytic_der_couple_phase(sim, dh_qc_dzc, eigvals, eigvecs):
    """
    Calculates the phase change needed to fix the gauge using analytical derivative
    couplings, i.e., calculates the phase-factors :math:`u^{q}_{i}` and :math:`u^{p}_{i}`
    such that :math:`d_{ij}^{q}u_{i}^{q*}u_{j}^{q}` and :math:`d_{ij}^{p}u_{i}^{p*}u_{j}^{p}`
    are real-valued.

    It does this by calculating the derivative couplings analytically. In the event of
    degenerate eigenvalues, an error is logged and a small offset is added to the energy
    differences.

    .. rubric:: Args
    sim: Simulation
        Simulation object.
    dh_qc_dzc : tuple
        Sparse representation of the derivative of the quantum-classical Hamiltonian
        with respect to the conjugate complex coordinate.
    eigvals : ndarray
        Eigenvalues of the quantum subsystem.
    eigvecs : ndarray
        Eigenvectors of the quantum subsystem.

    .. rubric::
    der_couple_dq_phase : ndarray
        Phase factor for derivative couplings obtained by differentiating
        w.r.t. the position coordinate.
    der_couple_dp_phase : ndarray
        Phase factor for derivative couplings obtained by differentiating
        w.r.t. the momentum coordinate.
    """
    inds, mels, shape = dh_qc_dzc
    batch_size = shape[0]
    m = sim.model.constants.classical_coordinate_mass
    h = sim.model.constants.classical_coordinate_weight
    num_classical_coords = shape[1]
    num_quantum_states = shape[2]
    der_couple_dq_phase = np.ones(
        (
            batch_size,
            num_quantum_states,
        ),
        dtype=complex,
    )
    der_couple_dp_phase = np.ones(
        (
            batch_size,
            num_quantum_states,
        ),
        dtype=complex,
    )
    for i in range(num_quantum_states - 1):
        j = i + 1
        evec_i = eigvecs[..., i]
        evec_j = eigvecs[..., j]
        eval_i = eigvals[..., i]
        eval_j = eigvals[..., j]
        eigval_diff = eval_j - eval_i
        plus = np.zeros_like(eigval_diff)
        if sim.settings.debug:
            if np.any(np.abs(eigval_diff) < numerical_constants.SMALL):
                plus[np.where(np.abs(eigval_diff) < numerical_constants.SMALL)] = 1
                logger.error("Degenerate eigenvalues detected.")
        der_couple_dzc = np.zeros(
            (
                batch_size,
                num_classical_coords,
            ),
            dtype=complex,
        )
        der_couple_dz = np.zeros(
            (
                batch_size,
                num_classical_coords,
            ),
            dtype=complex,
        )
        np.add.at(
            der_couple_dzc,
            (inds[0], inds[1]),
            np.conj(evec_i)[inds[0], inds[2]]
            * mels
            * evec_j[inds[0], inds[3]]
            / ((eigval_diff + plus)[inds[0]]),
        )
        np.add.at(
            der_couple_dz,
            (inds[0], inds[1]),
            np.conj(evec_i)[inds[0], inds[3]]
            * np.conj(mels)
            * evec_j[inds[0], inds[2]]
            / ((eigval_diff + plus)[inds[0]]),
        )
        der_couple_dq, der_couple_dp = dzdzc_to_dqdp(
            der_couple_dz, der_couple_dzc, m[np.newaxis, :], h[np.newaxis, :]
        )
        der_couple_dq_angle = np.angle(
            der_couple_dq[
                np.arange(len(der_couple_dq)),
                np.argmax(np.abs(der_couple_dq), axis=-1),
            ]
        )
        der_couple_dp_angle = np.angle(
            der_couple_dp[
                np.arange(len(der_couple_dp)),
                np.argmax(np.abs(der_couple_dp), axis=-1),
            ]
        )
        der_couple_dq_angle[
            np.where(np.abs(der_couple_dq_angle) < numerical_constants.SMALL)
        ] = 0
        der_couple_dp_angle[
            np.where(np.abs(der_couple_dp_angle) < numerical_constants.SMALL)
        ] = 0
        der_couple_dq_phase[..., i + 1 :] = (
            np.exp(1j * der_couple_dq_angle[..., np.newaxis])
            * der_couple_dq_phase[..., i + 1 :]
        )
        der_couple_dp_phase[..., i + 1 :] = (
            np.exp(1j * der_couple_dp_angle[..., np.newaxis])
            * der_couple_dp_phase[..., i + 1 :]
        )
    return der_couple_dq_phase, der_couple_dp_phase


def calc_resc_dir_z_fssh(
    sim, eigval_diff, eigvec_init_state, eigvec_final_state, dh_qc_dzc_traj
):
    """
    Calculates the rescaling direction for FSSH using analytical derivative couplings.

    This function is not vectorized over multiple trajectories and is intended to be
    only called when needed in that trajectory.

    It calculates both the derivative coupling w.r.t. z and z*, and checks that they
    are properly aligned to correspond with real-valued phase space derivative couplings.
    If they are not, an error is logged.

    .. rubric:: Args
    eigval_diff : ndarray
        Difference in eigenvalues between the initial and final states, ``e_final - e_initial``.
    eigvec_init_state : ndarray
        Eigenvector of the initial state.
    eigvec_final_state : ndarray
        Eigenvector of the final state.
    dh_qc_dzc_traj : tuple
        Sparse representation of the derivative of the quantum-classical Hamiltonian
        with respect to the conjugate complex coordinate.
    m : ndarray
        Classical coordinate mass.
    h : ndarray
        Classical coordinate weight.

    .. rubric:: Returns
    dkj_zc : ndarray
        Nonadiabatic coupling vector for rescaling the z coordinate.
        Derivative is w.r.t. z*.
    """
    inds, mels, shape = dh_qc_dzc_traj
    num_classical_coordinates = shape[1]
    dkj_z = np.zeros((num_classical_coordinates), dtype=complex)
    dkj_zc = np.zeros((num_classical_coordinates), dtype=complex)
    np.add.at(
        dkj_zc,
        (inds[1]),
        np.conj(eigvec_init_state)[inds[2]]
        * mels
        * eigvec_final_state[inds[3]]
        / eigval_diff,
    )
    if sim.settings.debug:
        np.add.at(
            dkj_z,
            (inds[1]),
            np.conj(eigvec_init_state)[inds[3]]
            * np.conj(mels)
            * eigvec_final_state[inds[2]]
            / eigval_diff,
        )
        # Determine the position of the maximum derivative coupling.
        max_pos = np.argmax(np.abs(dkj_zc))
        dkj_zc_max = dkj_zc[max_pos]
        dkj_z_max = dkj_z[max_pos]
        # Calculate the fraction of error in the real and imaginary parts,
        # i.e., the deviation from dkj_zc = dkj_z*.
        alignment_error = np.abs(dkj_zc_max - np.conj(dkj_z_max)) / np.abs(dkj_zc_max)
        if alignment_error > numerical_constants.GAUGE_FIX_THRESHOLD:
            logger.warning(
                "Derivative couplings are not properly aligned. "
                "Gauge fixing may be needed.\n"
                "Fraction of alignment error: %s",
                alignment_error,
            )

    return dkj_zc


def numerical_fssh_hop(model, parameters, **kwargs):
    """
    Determines the shift required to conserve energy during a hop in FSSH using
    an iterative numerical method. The coordinate following the hop is ``z + shift``.

    The algorithm is as follows:
    1. Calculate the initial energy using the Hamiltonian function at the current ``z``.
    2. Define a grid from ``-gamma_range`` to ``+gamma_range`` with ``num_points`` points
       uniformly spaced.
    3. Calculate the energy at each point in the grid using the Hamiltonian function.
    4. Find the point in the grid that minimizes the difference between the energy
       difference and the calculated energy difference.
    5. Recenter the grid around the minimum point found in step 4, reduce the
       ``gamma_range`` by half, and repeat steps 3-5 until either the minimum energy
       difference is less than the threshold or the maximum number of iterations
       is reached.
    6. If the minimum energy difference is less than the threshold, return the
       corresponding shift. Otherwise, return a zero shift and indicate that the
       hop was frustrated.

    .. rubric:: Keyword Arguments
    z : ndarray
        Current complex coordinate.
    resc_dir_z : float
        Rescaling direction.
    eigval_diff : float
        Difference in eigenvalues between the initial and final states, ``e_final - e_initial``.


    .. rubric:: Model Constants
    numerical_fssh_hop_gamma_range : float, default: 5.0
        Initial range (negative to positive) of gamma values to search over.
    numerical_fssh_hop_max_iter : int, default: 20
        Maximum number of iterations to perform.
    numerical_fssh_hop_num_points : int, default: 10
        Number of points to sample in each iteration.
    numerical_fssh_hop_threshold : float, default: 1e-6
        Energy threshold for convergence.

    .. rubric:: Returns
    ndarray
        The shift to apply to the complex coordinate to conserve energy.
    hop_successful : bool
        Whether the hop was successful (i.e., energy conservation was achieved).
    """
    z = kwargs["z"]
    resc_dir_z = kwargs["resc_dir_z"]
    eigval_diff = kwargs["eigval_diff"]
    gamma_range = model.constants.get("numerical_fssh_hop_gamma_range", 5.0)
    max_iter = model.constants.get("numerical_fssh_hop_max_iter", 20)
    num_points = model.constants.get("numerical_fssh_hop_num_points", 10)
    thresh = model.constants.get("numerical_fssh_hop_threshold", 1e-6)
    h_c, _ = model.get("h_c")
    init_energy = h_c(model, parameters, z=np.array([z]), batch_size=1)[0]
    min_gamma = 0.0
    num_iter = 0
    min_energy = 1.0
    while min_energy > thresh and num_iter < max_iter:
        gamma_list = np.linspace(
            min_gamma - gamma_range, min_gamma + gamma_range, num_points
        )
        new_energies = np.abs(
            eigval_diff
            - np.array(
                [
                    init_energy
                    - h_c(
                        model,
                        parameters,
                        z=np.array([z - 1j * gamma * resc_dir_z]),
                        batch_size=1,
                    )[0]
                    for gamma in gamma_list
                ]
            )
        )
        min_gamma = gamma_list[np.argmin(new_energies)]
        min_energy = np.min(new_energies)
        gamma_range = gamma_range * 0.5
        num_iter += 1
    if min_energy > thresh:
        return np.zeros_like(z), False
    return -1j * min_gamma * resc_dir_z, True



"""
This module contains the MeanField algorithm class.
"""

from functools import partial
from qclab.algorithm import Algorithm
from qclab import tasks


class MeanField(Algorithm):
    """
    Mean-field dynamics algorithm class.
    """

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        self.default_settings = {}
        super().__init__(self.default_settings, settings)

    initialization_recipe = [
        tasks.initialize_variable_objects,
        tasks.initialize_norm_factor,
        tasks.initialize_z,
        tasks.update_h_q_tot,
    ]

    update_recipe = [
        # Begin RK4 integration steps.
        partial(tasks.update_classical_force, z_name="z"),
        tasks.update_quantum_classical_force,
        tasks.update_z_rk4_k123,
        partial(tasks.update_classical_force, z_name="z_1"),
        partial(
            tasks.update_quantum_classical_force,
            z_name="z_1",
            wf_changed=False,
        ),
        partial(tasks.update_z_rk4_k123, z_name="z", z_k_name="z_2", k_name="z_rk4_k2"),
        partial(tasks.update_classical_force, z_name="z_2"),
        partial(
            tasks.update_quantum_classical_force,
            z_name="z_2",
            wf_changed=False,
        ),
        partial(
            tasks.update_z_rk4_k123,
            z_name="z",
            z_k_name="z_3",
            k_name="z_rk4_k3",
            dt_factor=1.0,
        ),
        partial(tasks.update_classical_force, z_name="z_3"),
        partial(
            tasks.update_quantum_classical_force,
            z_name="z_3",
            wf_changed=False,
        ),
        tasks.update_z_rk4_k4,
        # End RK4 integration steps.
        tasks.update_wf_db_rk4,
        tasks.update_h_q_tot,
    ]

    collect_recipe = [
        tasks.update_t,
        tasks.update_dm_db_wf,
        tasks.update_quantum_energy_wf,
        tasks.update_classical_energy,
        tasks.collect_t,
        tasks.collect_dm_db,
        tasks.collect_classical_energy,
        tasks.collect_quantum_energy,
    ]


class MeanFieldAbInitio(Algorithm):
    """
    Mean-field dynamics algorithm class implemented in the adiabatic basis
    and compatible with ab initio calculations.
    """

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        self.default_settings = {
            "update_wf_adb_eig_num_substeps": 10,
            "use_wf_overlaps_for_adb_connection": False,
        }
        super().__init__(self.default_settings, settings)

    initialization_recipe = [
        tasks.initialize_variable_objects,
        partial(tasks.copy_to_parameters, state_name="seed", parameters_name="seed"),
        tasks.initialize_norm_factor,
        tasks.initialize_z,
        partial(
            tasks.update_ab_initio_property,
            property_dict={
                "energy": {"z": "z", "excited_amplitudes": True},
                "gradient": {"z": "z", "state_inds_gradient": None},
                "derivative_coupling": {
                    "z": "z",
                    "state_inds_derivative_coupling": None,
                },
            },
        ),
        tasks.update_h_q_tot,
        tasks.update_classical_force,
        tasks.update_derivative_coupling_dzc,
        partial(tasks.update_quantum_classical_force, wf_db_name="wf_adb"),
        tasks.update_adb_connection,
    ]

    update_recipe = [
        partial(
            tasks.copy_in_state,
            copy_name="aip_excited_amplitudes_previous",
            orig_name="aip_excited_amplitudes",
        ),
        partial(
            tasks.copy_in_state,
            copy_name="adb_connection_previous",
            orig_name="adb_connection",
        ),
        partial(tasks.copy_in_state, copy_name="h_q_tot_previous", orig_name="h_q_tot"),
        partial(
            tasks.copy_in_state,
            copy_name="quantum_classical_force_previous",
            orig_name="quantum_classical_force",
        ),
        partial(
            tasks.copy_in_state,
            copy_name="classical_force_previous",
            orig_name="classical_force",
        ),
        partial(
            tasks.copy_in_state,
            copy_name="z_previous",
            orig_name="z",
        ),
        tasks.update_q_velocity_verlet,
        partial(
            tasks.update_ab_initio_property,
            property_dict={
                "energy": {"z": "z", "excited_amplitudes": True},
                "gradient": {"z": "z", "state_inds_gradient": None},
                "derivative_coupling": {
                    "z": "z",
                    "state_inds_derivative_coupling": None,
                },
                "wf_overlaps": {
                    "z": "z",
                    "z_previous": "z_previous",
                    "amplitudes_previous": "aip_excited_amplitudes_previous",
                    "amplitudes_current": "aip_excited_amplitudes",
                },
            },
        ),
        tasks.update_derivative_coupling_dzc,
        tasks.update_derivative_coupling_dzc_gauge,
        tasks.update_wf_overlaps_gauge,
        partial(tasks.update_adb_connection, update_derivative_coupling=False),
        tasks.update_h_q_tot,
        partial(
            tasks.update_quantum_classical_force,
            wf_db_name="wf_adb",
        ),
        partial(
            tasks.update_wf_adb_hop_prob,
            update_hopping_probabilities=False,
        ),
        tasks.update_p_velocity_verlet,
        tasks.update_classical_force,
    ]

    collect_recipe = [
        tasks.update_t,
        partial(tasks.update_dm_db_wf, wf_db_name="wf_adb"),
        partial(tasks.update_quantum_energy_wf, wf_db_name="wf_adb"),
        tasks.update_classical_energy,
        tasks.collect_t,
        partial(tasks.collect_dm_db, dm_db_name="dm_db", dm_db_output_name="dm_adb"),
        tasks.collect_classical_energy,
        tasks.collect_quantum_energy,
    ]
